//=========================================================================
//Copyright LizardOfOz.
//This is a VERY early version intended to prototype maps.
//Expect things to change over time.
//
//Check the following links for updates:
//   https://tf2maps.net/members/lizardofoz.39426/#resources
//   https://twitter.com/LizardOfOz_
//   https://youtube.com/LizardOfOz
//
//Credits:
//  LizardOfOz - the original VSH Plugin. VSH Vscript recreation.
//  Maxxy, Velly, JPRAS - Saxton Hale's model imitating the Jungle Inferno short.
//  MegapiemanPHD - Hale's voice lines.
//  Yakibomb - give_tf_weapon script bundle (used for Hale's first-person hands model).
//=========================================================================

haleStunBreakoutAbility <- null;
haleSlamAbility <- null;
haleChargeAbility <- null;
haleSaxtonPunchAbility <- null;
haleBraveJumpAbility <- null;
haleAbilities <- [];

function TickHaleAbilities(void)
{
    if (IsNotValidRound() || IsRoundOver())
        return;
    local hale = GetHale();
    if (IsNotHale(hale))
        return;

    foreach (ability in haleAbilities)
        ability.Tick(hale, 0.1);
}
AddListener("tick", TickHaleAbilities, 0);

function FrameTickHaleAbilities(void)
{
    if (IsNotValidRound() || IsRoundOver())
        return;
    local hale = GetHale();
    if (IsNotHale(hale))
        return;

    foreach (ability in haleAbilities)
        ability.FrameTick(hale);
}
AddListener("frame_tick", FrameTickHaleAbilities, 0);

class HaleAbility
{
    meter = 0;

    function Tick(hale, interval) { }
    function FrameTick(hale) { }
    function IsReady(hale) { return false; }
    function Perform(hale) { }
    function MeterValueAsString() { return null; }
}

class StunBreakOutHaleAbility extends HaleAbility
{
    haleLastTimeHolidayPunch = 0;

    function Tick(hale, interval)
    {
        if (hale.InCond(Constants.ETFCond.TF_COND_STUNNED) || ((Time() - haleLastTimeHolidayPunch <= 4) && hale.InCond(Constants.ETFCond.TF_COND_TAUNTING)))
            meter += interval;
        else if (meter > 0)
            meter -= interval;
        if (meter < 0)
            meter = 0;

        if (meter >= 4)
            Perform(hale);
    }

    function Perform(hale)
    {
        meter = 0;
        EmitSoundOn("VSH.Rage", hale);
        hale.RemoveCond(Constants.ETFCond.TF_COND_STUNNED);
        hale.RemoveCond(Constants.ETFCond.TF_COND_TAUNTING);
        NetProps.SetPropEntity(hale, "m_hGroundEntity", null);
        hale.ApplyAbsVelocityImpulse(Vector(0,0,800));

        DispatchParticleEffect("hammer_impact_button", hale.GetOrigin() + Vector(0,0,20), Vector(0,0,0));
        EmitSoundOn("VSH.Slam.Impact", hale);

        vsh_vscript.CreateAoE(hale.GetCenter(), 400,
            function (target, deltaVector, distance) {
                target.TakeDamageEx(
                    hale,
                    hale,
                    hale.GetActiveWeapon(),
                    deltaVector,
                    target.GetCenter(),
                    1,
                    1);
            }
            function (target, deltaVector, distance) {
                local pushForce = distance < 100 ? 10 : 10 / sqrt(distance);
                NetProps.SetPropEntity(target, "m_hGroundEntity", null);
                deltaVector.x = deltaVector.x * 1450 * pushForce;
                deltaVector.y = deltaVector.y * 1450 * pushForce;
                deltaVector.z = 450 * pushForce;
                target.ApplyAbsVelocityImpulse(deltaVector);
            });
    }
}

class SaxtonPunchHaleAbility extends HaleAbility
{
    meter = -30;
    function IsReady(hale)
    {
        return meter == 0;
    }

    function MeterValueAsString()
    {
        if (meter == 0)
            return "Ready";
        else if (meter < 0)
            return format("%.1fs", -meter);
        else
            return format("%d%%", meter * 100);
    }

    function Tick(hale, interval)
    {
        if (meter < 0 && !vsh_vscript.IsRoundSetup())
        {
            meter += interval;
            if (meter >= 0)
            {
                meter = 0;
                EmitSoundOn("VSH.PunchReadyVO", hale);
                EmitSoundOnClient("VSH.PunchReadySFX", hale);
                NetProps.SetPropEntity(hale, "m_hActiveWeapon", hale.GiveWeapon("Hale's Own Crits"));
            }
        }
    }

    function Perform(hale, victim)
    {
        if (!IsReady(hale))
            return;
        meter -= 30;

        local haleEyeVector = hale.EyeAngles().Forward();
        haleEyeVector.Norm();

        NetProps.SetPropEntity(hale, "m_hActiveWeapon", hale.GiveWeapon("Hale's Own Fists"));
        EmitSoundOn("TFPlayer.CritHit", hale);
        EmitSoundOn("VSH.PunchSFX", hale);
        if (RandomInt(1,3))
            EmitSoundOn("VSH.Punch", hale);
        DispatchParticleEffect("passtime_air_blast", victim.GetCenter(), haleEyeVector);
        DispatchParticleEffect("Explosion_ShockWave_01", victim.GetOrigin(), Vector(0,0,0));
        ScreenShake(hale.GetCenter(), 10, 2.5, 1, 1000, 0, true);

        vsh_vscript.CreateAoE(hale.GetCenter(), 600,
            function (target, deltaVector, distance) {
                local dot = haleEyeVector.Dot(deltaVector);
                if (dot < 0.6)
                    return;
                DispatchParticleEffect("mvm_pow_punch", target.EyePosition() + Vector(0,0,30), Vector(0,0,0));
                target.TakeDamageEx(
                    hale,
                    hale,
                    hale.GetActiveWeapon(),
                    deltaVector,
                    target.GetCenter(),
                    target.GetMaxHealth() * (0.6 - distance / 1000),
                    1);
            }
            function (target, deltaVector, distance) {
                local dot = haleEyeVector.Dot(deltaVector);
                if (dot < 0.6)
                    return;
                local pushForce = distance < 100 ? 10 : 10 / sqrt(distance);
                NetProps.SetPropEntity(target, "m_hGroundEntity", null);
                deltaVector.x = deltaVector.x * 1250 * pushForce;
                deltaVector.y = deltaVector.y * 1250 * pushForce;
                deltaVector.z = 750 * pushForce;
                target.ApplyAbsVelocityImpulse(deltaVector);
            });
    }
}

class SlamHaleAbility extends HaleAbility
{
    inUse = false;
    lastFrameDownVelocity = 0;

    function IsReady(hale)
    {
        return meter == 0;
    }

    function MeterValueAsString()
    {
        if (meter == 0)
            return "Ready";
        else if (meter < 0)
            return format("%.1fs", -meter);
        else
            return format("%d%%", meter * 100);
    }

    function Tick(hale, interval)
    {
        if (meter < 0)
        {
            meter += interval;
            if (meter > 0)
                meter = 0;
        }
        if (vsh_vscript.haleChargeAbility.meter > 0 || meter < 0)
            return;
        if (!(hale.GetFlags() & Constants.FPlayer.FL_ONGROUND) && (hale.GetFlags() & Constants.FPlayer.FL_DUCKING))
            Weightdown(hale);
        else if (inUse && !(hale.GetFlags() & Constants.FPlayer.FL_DUCKING))
        {
            inUse = false;
            hale.RemoveCond(Constants.ETFCond.TF_COND_HALLOWEEN_KART_DASH);
            hale.SetGravity(1);
        }
        else if (inUse && (hale.GetFlags() & Constants.FPlayer.FL_ONGROUND))
        {
            inUse = false;
            hale.RemoveCond(Constants.ETFCond.TF_COND_HALLOWEEN_KART_DASH);
            hale.SetGravity(1);
            if (lastFrameDownVelocity < -700)
                Perform(hale);
        }
    }

    function Weightdown(hale)
    {
        inUse = true;
        hale.SetGravity(3);
        lastFrameDownVelocity = hale.GetAbsVelocity().z;
        if (lastFrameDownVelocity < -700)
            hale.AddCond(Constants.ETFCond.TF_COND_HALLOWEEN_KART_DASH);
    }

    function Perform(hale)
    {
        DispatchParticleEffect("hammer_impact_button", hale.GetOrigin() + Vector(0,0,20), Vector(0,0,0));
        EmitSoundOn("VSH.Slam.Impact", hale);
        lastFrameDownVelocity = 0;
        meter = -5;

        local currentWeapon = hale.GetActiveWeapon();
        local mannthreadsWeapon = NetProps.GetPropEntityArray(hale, "m_hMyWeapons", 1);
        NetProps.SetPropEntity(hale, "m_hActiveWeapon", mannthreadsWeapon);

        vsh_vscript.CreateAoE(hale.GetCenter(), 600,
            function (target, deltaVector, distance) {
                target.TakeDamageEx(
                    hale,
                    hale,
                    hale.GetActiveWeapon(),
                    deltaVector,
                    target.GetCenter(),
                    target.GetMaxHealth() * (1 - distance / 600),
                    1);
            }
            function (target, deltaVector, distance) {
                local pushForce = distance < 100 ? 1 : 100 / distance;
                NetProps.SetPropEntity(target, "m_hGroundEntity", null);
                deltaVector.x = deltaVector.x * 1250 * pushForce;
                deltaVector.y = deltaVector.y * 1250 * pushForce;
                deltaVector.z = 950 * pushForce;
                target.ApplyAbsVelocityImpulse(deltaVector);
            });

        NetProps.SetPropEntity(hale, "m_hActiveWeapon", currentWeapon);
        ScreenShake(hale.GetCenter(), 10, 2.5, 1, 1000, 0, true);
    }
}

class ChargeHaleAbility extends HaleAbility
{
    isCurrentlyDashing = false;
    haleLastFrameDownVelocity = 0;
    voiceTime = 0;
    voiceRNG = 0;
    bashedByHale = [];
    midAirWindUpOverload = 0;

    function IsReady(hale)
    {
        return meter == 0;
    }

    function MeterValueAsString()
    {
        if (meter == 0)
            return "Ready";
        else if (meter < 0)
            return format("%.1fs", -meter);
        else
            return format("%d%%", meter * 100);
    }

    function Tick(hale, interval)
    {
        if (isCurrentlyDashing)
            return;

        if (meter < 0)
        {
            meter += interval;
            if (meter > 0)
            {
                EmitSoundOnClient("TFPlayer.ReCharged", hale);
                meter = 0;
            }
            else
                return;
        }
    }

    function FrameTick(hale)
    {
        if (isCurrentlyDashing)
        {
            TickDash(hale);
            return;
        }
        if (meter < 0)
            return;

        local buttons = NetProps.GetPropInt(hale, "m_nButtons");
        if (hale.GetTimeSinceCalledForMedic() < 0.2
            || hale.IsUsingActionSlot()
            || (buttons & Constants.FButtons.IN_ATTACK3)
            || (buttons & Constants.FButtons.IN_RELOAD))
        {
            WindUp(hale);
        }
        else if (meter >= 0.01)
            Perform(hale);
    }

    function WindUp(hale)
    {
        meter += 0.01;
        if (meter >= 1.0)
        {
            meter = 1.0;
            if (!(hale.GetFlags() & Constants.FPlayer.FL_ONGROUND))
            {
                midAirWindUpOverload += 0.01;
                if (midAirWindUpOverload >= 1)
                {
                    Perform(hale);
                    return;
                }
            }
            else
                midAirWindUpOverload = 0;
        }
        local main_viewmodel = NetProps.GetPropEntity(hale, "m_hViewModel")
        local sequenceId = main_viewmodel.LookupSequence("grapple_pull_idle");
        if (main_viewmodel.GetSequence() != sequenceId)
            main_viewmodel.SetSequence(sequenceId)
        hale.AddCustomAttribute("no_attack", 0.25, 1);
        if (Time() > voiceTime)
        {
            voiceRNG = RandomInt(1,3)
            voiceTime = Time() + 999
            EmitSoundOn("VSH.Charge"+voiceRNG, hale);
            EmitSoundOn("VSH.ChargeSFX", hale);
        }
        hale.SetAbsVelocity(Vector(0,0,0));
        hale.AddCustomAttribute("move speed penalty", 0.25, 1);
    }

    function Perform(hale)
    {
        isCurrentlyDashing = true;
        bashedByHale = [];

        DoEntFire("hale_charge_zone", "Enable", "", 0, hale, hale)
        DoEntFire("hale_charge_catapult", "Enable", "", 0, hale, hale)

        local chargeDuration = meter / 1.66 + 0.4;

        NetProps.SetPropEntity(hale, "m_hGroundEntity", null);
        local forward = hale.EyeAngles();
        local pitch = hale.GetFlags() & Constants.FPlayer.FL_ONGROUND ? forward.Pitch() - 10 : forward.Pitch();
        forward = QAngle(pitch, forward.Yaw(), 0);
        hale.ApplyAbsVelocityImpulse(forward.Forward() * 800 * chargeDuration)
        hale.SetGravity(0.3);
        voiceTime = 0;
        midAirWindUpOverload = 0;

        hale.AddCondEx(100, chargeDuration, null);
        hale.AddCondEx(Constants.ETFCond.TF_COND_SHIELD_CHARGE, chargeDuration, null);
        hale.AddCondEx(Constants.ETFCond.TF_COND_KNOCKED_INTO_AIR, chargeDuration, null);
        hale.AddCustomAttribute("no_attack", 1, chargeDuration);
        EmitSoundOn("VSH.Dash" + voiceRNG, hale);
        EmitSoundOn("VSH.DashSFX", hale);
        RunWithDelay("haleChargeAbility.Finish(activator)", hale, chargeDuration);
    }

    function Finish(hale)
    {
        meter = -10;
        isCurrentlyDashing = false;
        //local velocity = hale.GetAbsVelocity();
        hale.SetGravity(1);
        //hale.SetAbsVelocity(Vector(velocity.x/3,velocity.y/3,velocity.z/3));
        DoEntFire("hale_charge_zone", "Disable", "", 0, hale, hale)
        DoEntFire("hale_charge_catapult", "Disable", "", 0, hale, hale)
    }

    function TickDash(hale)
    {
        local haleForwardDirection = hale.EyeAngles().Forward();
        //NetProps.SetPropEntity(hale, "m_hGroundEntity", null);
        NetProps.SetPropInt(hale, "m_Shared.m_bJumping", 1);
        //NetProps.SetPropInt(hale, "m_Shared.m_iAirDash", 1);
        NetProps.SetPropFloat(hale, "m_Shared.m_flJumpTime", 1);

        NetProps.SetPropEntity(hale, "m_hGroundEntity", null);
        //local force = (hale.GetFlags() & Constants.FPlayer.FL_ONGROUND ? 3000 : 1000);
        local force = 1400;
        hale.SetAbsVelocity(haleForwardDirection*force)

        local catapult = Entities.FindByName(null, "hale_charge_catapult");
        catapult.SetAbsOrigin(hale.GetCenter());

        local forwardOffset = haleForwardDirection * 60;
        local chargeZone = Entities.FindByName(null, "hale_charge_zone");
        chargeZone.SetAbsOrigin(hale.GetCenter() + forwardOffset);

        local main_viewmodel = NetProps.GetPropEntity(hale, "m_hViewModel")
        main_viewmodel.SetSequence(main_viewmodel.LookupSequence("grapple_pull_idle") )
        hale.SetSequence(hale.LookupSequence("primary_floatA") )

        local building = null;
        while (building = Entities.FindByClassname(building, "obj_*"))
        {
            if ((hale.GetCenter() - building.GetOrigin()).LengthSqr() < 10000)
            building.TakeDamageEx(hale, hale, hale.GetActiveWeapon(), Vector(0,0,0), building.GetCenter(), 100, 1);
        }
    }

    function BashTarget(victim)
    {
        local hale = GetHale();
        printl("bash1 "+victim)
        if (IsNotHale(hale) || IsHale(victim) || bashedByHale.find(victim) != null)
            return;
        printl("bash2 "+victim)

        bashedByHale.push(victim);
        EmitSoundOn("DemoCharge.HitFlesh", victim);

        NetProps.SetPropEntity(victim, "m_hGroundEntity", null);
        local victimVelocity = victim.GetAbsVelocity();
        if (victimVelocity.z < 0)
            victimVelocity.z = 0;

        local forward = victim.GetCenter() - hale.GetCenter();
        forward.Norm();
        forward.z = 350;
        forward.x = forward.x * 1250
        forward.y = forward.y * 1250
        victim.SetAbsVelocity(victimVelocity + forward);

        victim.TakeDamageEx(hale, hale, hale.GetActiveWeapon(), forward, victim.GetCenter(), victim.GetMaxHealth() * 0.55, 1);
    }
}

class BraveJumpHaleAbility extends HaleAbility
{
    haleJumpStatus = HALE_JUMP_STATUS.WALKING;

    function FrameTick(hale)
    {
        local buttons = NetProps.GetPropInt(hale, "m_nButtons");

        if (!(hale.GetFlags() & Constants.FPlayer.FL_ONGROUND))
        {
            if (haleJumpStatus == HALE_JUMP_STATUS.WALKING)
                haleJumpStatus = HALE_JUMP_STATUS.JUMP_STARTED;
            else if (haleJumpStatus == HALE_JUMP_STATUS.JUMP_STARTED && !(buttons & Constants.FButtons.IN_JUMP))
                haleJumpStatus = HALE_JUMP_STATUS.CAN_DOUBLE_JUMP;
        }
        else
            haleJumpStatus = HALE_JUMP_STATUS.WALKING;

        if (buttons & Constants.FButtons.IN_JUMP && haleJumpStatus == HALE_JUMP_STATUS.CAN_DOUBLE_JUMP)
        {
            //PlayVOAtEntity(hale, SOUND_VO_SAXTON_JUMP);
            if (!vsh_vscript.IsRoundSetup())
                EmitSoundOn("VSH.Jump", hale);

            haleJumpStatus = HALE_JUMP_STATUS.DOUBLE_JUMPED;
            Perform(hale);
        }
    }

    function Perform(hale)
    {
        local buttons = NetProps.GetPropInt(hale, "m_nButtons");
        local eyeAngles = hale.EyeAngles();
        local forward = eyeAngles.Forward();
        forward.z = 0;
        forward.Norm();
        local left = eyeAngles.Left();
        left.z = 0;
        left.Norm();

        local forwardmove = 0
        if (buttons & Constants.FButtons.IN_FORWARD)
            forwardmove = 1;
        else if (buttons & Constants.FButtons.IN_BACK)
            forwardmove = -1;
        local sidemove = 0
        if (buttons & Constants.FButtons.IN_MOVELEFT)
            sidemove = -1;
        else if (buttons & Constants.FButtons.IN_MOVERIGHT)
            sidemove = 1;

        local newVelocity = Vector(0,0,0);
        newVelocity.x = forward.x * forwardmove + left.x * sidemove;
        newVelocity.y = forward.y * forwardmove + left.y * sidemove;
        newVelocity.Norm();
        newVelocity *= 300;
        newVelocity.z = vsh_vscript.haleJumpForce

        local currentVelocity = hale.GetAbsVelocity();
        if (currentVelocity.z < 0)
            currentVelocity.z = 0;

        NetProps.SetPropEntity(hale, "m_hGroundEntity", null);
        hale.SetAbsVelocity(currentVelocity + newVelocity)
    }}

function CreateAoE(center, radius, applyDamageFunc, applyPushFunc)
{
    foreach(target in GetAliveMercs())
    {
        local deltaVector = target.GetCenter() - center;
        local distance = deltaVector.Norm();
        if (distance > radius)
            continue;

        applyPushFunc(target, deltaVector, distance);
        applyDamageFunc(target, deltaVector, distance);
    }

    local target = null;
    while (target = Entities.FindByClassname(target, "obj_*"))
    {
        local deltaVector = target.GetCenter() - center;
        local distance = deltaVector.Norm();
        if (distance > radius)
            continue;

        applyDamageFunc(target, deltaVector, distance);
    }
}

haleStunBreakoutAbility <- StunBreakOutHaleAbility();
haleSlamAbility <- SlamHaleAbility();
haleChargeAbility <- ChargeHaleAbility();
haleSaxtonPunchAbility <- SaxtonPunchHaleAbility();
haleBraveJumpAbility <- BraveJumpHaleAbility();
haleAbilities <- [haleStunBreakoutAbility, haleSlamAbility, haleChargeAbility, haleSaxtonPunchAbility, haleBraveJumpAbility];
